﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace IK
{
    public delegate void ChainChangedHandler(object sender, EventArgs e);

    class Canvas : Control
    {
        public event ChainChangedHandler ChainChanged;

        private Xform xform = new Xform();
        private float scale = 1;

        private Chain ccd = null;
        private Link[] links = null;

        private Vector origin = new Vector(0, 0);
        private Vector target = new Vector(0.9, 0.01);


        private bool showAngleRange;

        public bool ShowAngleRange
        {
            get { return this.showAngleRange; }
            set { this.showAngleRange = value; }
        }




        public Canvas()
        {
            SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
            SetStyle(ControlStyles.ResizeRedraw, true);
            SetStyle(ControlStyles.UserPaint, true);
            UpdateStyles();
        }



        public void SetCCD(Chain ccd, Link[] links)
        {
            this.ccd = ccd;
            this.links = links;

            UpdateXform();
        }



        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            UpdateXform();
        }

        private void UpdateXform()
        {
            scale = 1;
            if (ccd != null)
                scale = (float)((links.Sum(f=>f.Length)) * 2.0 / (Math.Min(Width, Height)));

            xform.Update(Width, Height, scale);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            RectangleF circle = RectangleF.Empty;
            SolidBrush brush = new SolidBrush(Color.Empty);
            Pen pen = new Pen(Color.Empty);

            e.Graphics.SetClip(new Rectangle(1, 1, Width - 2, Height - 2));
            e.Graphics.Transform = xform.Transform;
            e.Graphics.SmoothingMode = SmoothingMode.AntiAlias;




            // Draw the axis
            pen.Color = Color.FromArgb(0xF0, 0xF0, 0xF0);
            e.Graphics.DrawLine(pen, -Width, (float)origin.Y, Width, (float)origin.Y);
            e.Graphics.DrawLine(pen, (float)origin.X, -Height, (float)origin.X, Height);

            if (ccd != null && links.Length > 0) {

                // Draw the origin
                circle = CircleBounds(origin, 5 * xform.Scale);
                e.Graphics.DrawEllipse(Pens.Black, circle);


                double datum = 0;
                Vector p0 = origin;
                Vector p1;

                brush.Color = Color.FromArgb(32, Color.Firebrick);
                pen.Color = Color.FromArgb(64, Color.Firebrick);

                for (int i = 0; i < links.Length; i++) {
                    p1 = p0 + Vector.FromPolar(links[i].Length/scale, datum + links[i].Angle);

                    // Draw the joint
                    circle = CircleBounds(p0, 3);
                    e.Graphics.FillEllipse(Brushes.Black, circle);
                    e.Graphics.DrawEllipse(Pens.Black, circle);


                    //if (showAngleRange) {

                    //    circle = CircleBounds(p0, links[i].Length);
                    //    float start = (float)Maths.ToDeg(datum - CCD.jointAngleConstraint);
                    //    float sweep = (float)Maths.ToDeg(CCD.jointAngleConstraint * 2.0);

                    //    e.Graphics.FillPie(brush, Rectangle.Truncate(circle), start, sweep);
                    //    e.Graphics.DrawArc(pen, Rectangle.Truncate(circle), start, sweep);

                    //    Vector p2 = p0 + Vector.FromPolar(ccd.Chain[i].Length, datum);
                    //    e.Graphics.DrawLine(pen, (PointF)p0, (PointF)p2);
                    //}

                    // Draw the joint
                    e.Graphics.DrawLine(Pens.Black, (PointF)p0, (PointF)p1);


                    datum += links[i].Angle;
                    p0 = p1;
                }

                // Draw the end effector
                circle = CircleBounds(p0, 3);
                e.Graphics.FillEllipse(Brushes.Red, circle);
                e.Graphics.DrawEllipse(Pens.Red, circle);


                // Draw the target
                circle = CircleBounds(target/scale, 5);
                e.Graphics.DrawEllipse(Pens.Blue, circle);

                e.Graphics.Transform = new Matrix();
                e.Graphics.DrawString(string.Format("Target: [{0:0.0}, {1:0.0}]", target.X, target.Y), SystemFonts.DefaultFont, Brushes.Black, 2, 2);

                for (int i = 0; i < links.Length; i++) {
                    // Draw the joint angle text
                    e.Graphics.DrawString(string.Format("Angle {0:0.0}°", Maths.ToDeg(links[i].Angle)), SystemFonts.DefaultFont, Brushes.Black, 2, 2 + SystemFonts.DefaultFont.Height * (i + 1));
                }

                brush.Dispose();
                pen.Dispose();
            }

        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            e.Graphics.Clear(Color.White);
            e.Graphics.DrawRectangle(SystemPens.WindowFrame, 0, 0, Width - 1, Height - 1);
        }

        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (e.Button == MouseButtons.Left) {
                target = xform.FromXform(e.Location);
                target *= scale;

                //ccd.Target = target;
                //ccd.Solve(links, target, 5, 0.01);

                if (ChainChanged != null)
                    ChainChanged(this, new EventArgs());

                Invalidate();
            }

            if (e.Button == MouseButtons.Right) {
                target = xform.FromXform(e.Location);

                Invalidate();
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            if (e.Button == MouseButtons.Left) {
                if (ChainChanged != null)
                    ChainChanged(this, new EventArgs());
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (e.Button == MouseButtons.Left) {
                target = xform.FromXform(e.Location);
                target *= scale;

                //ccd.Target = target;
                //ccd.Solve(links, target, 5, 0.01);

                Invalidate();
            }

            if (e.Button == MouseButtons.Right) {
                target = xform.FromXform(e.Location);

                Invalidate();
            }
        }


        private static RectangleF CircleBounds(Vector centre, double radius)
        {
            float diameter = (float)(radius * 2);
            return new RectangleF((float)(centre.X - radius), (float)(centre.Y - radius), diameter, diameter);
        }

        public void PaintToGraphics(Graphics graphics, Rectangle clipBounds)
        {
            PaintEventArgs e = new PaintEventArgs(graphics, clipBounds);
            OnPaintBackground(e);
            OnPaint(e);
        }


        private class Xform
        {
            private Matrix xform;
            private Matrix xformInv;
            private int width;
            private int height;
            private float scale;


            public Xform()
            {
                this.xform = new Matrix();
                this.xformInv = new Matrix();
                this.width = 0;
                this.height = 0;
                this.scale = 1;
            }


            public Matrix Transform
            {
                get { return xform; }
            }
            public Matrix InverseTransform
            {
                get { return xformInv; }
            }
            public int Width
            {
                get { return width; }
            }
            public int Height
            {
                get { return height; }
            }
            public float Scale
            {
                get { return scale; }
            }


            public void Update(int width, int height, float scale)
            {
                this.width = width;
                this.height = height;
                this.scale = scale;

                // - Origin at the centre of the canvas.
                // - +X is right
                // - +Y is up
                scale = 1;
                // Set the XForm
                xform.Reset();
                xform.Scale(1.0f / scale, -1.0f / scale);
                xform.Translate((width * scale) / 2.0f, -(height * scale) / 2.0f);

                // Set the Inverse XForm
                xformInv = xform.Clone() as Matrix;
                if (!xformInv.IsInvertible)
                    throw new Exception("Unable to invert transformation matrix.");
                xformInv.Invert();
            }

            public PointF ToXform(PointF p)
            {
                PointF[] point = new PointF[] { p };
                xform.TransformPoints(point);
                return point[0];
            }

            public PointF FromXform(PointF p)
            {
                PointF[] point = new PointF[] { p };
                xformInv.TransformPoints(point);
                return point[0];
            }
        }
    }
}
