﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.ComponentModel;

namespace IK
{
    class ImageBox : Control
    {
        private Image image;
        private RectangleF destRect, srcRect, dragRect;

        private Point p0, p1;

        public ImageBox()
        {
            SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
            SetStyle(ControlStyles.ResizeRedraw, true);
            SetStyle(ControlStyles.UserPaint, true);
            UpdateStyles();
        }


        [Browsable(false)]
        public Image Image
        {
            get { return image; }
            set
            {
                image = value;
                srcRect = new RectangleF(0, 0, image.Width, image.Height);
                Invalidate();
            }
        }


        protected override void OnPaint(PaintEventArgs e)
        {
            if (image != null) {

                // Set the image display interpolation mode
                e.Graphics.InterpolationMode = InterpolationMode.NearestNeighbor;

                // Compute the client internal bounds with the image aspect ratio
                destRect = ComputeInternalBounds(ClientRectangle, (float)image.Width / (float)image.Height);

                // Draw the image
                e.Graphics.DrawImage(image, destRect, srcRect, GraphicsUnit.Pixel);


                if (!p0.Equals(p1)) {

                    // 
                    RectangleF bounds1 = ComputeBounds(p0, p1);
                    e.Graphics.DrawRectangle(Pens.Gray, bounds1.X, bounds1.Y, bounds1.Width, bounds1.Height);

                    // 
                    RectangleF bounds2 = ComputeExternalBounds(bounds1, (float)ClientSize.Width / (float)ClientSize.Height);
                    e.Graphics.DrawRectangle(Pens.White, bounds2.X, bounds2.Y, bounds2.Width, bounds2.Height);

                    dragRect = bounds2;
                }
            }
        }


        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);

            if (e.Button == MouseButtons.Left && ModifierKeys == Keys.Control) {
                if (image != null && destRect.Width > 0 && destRect.Height > 0) {
                    if (destRect.Contains(e.Location)) {
                        int px = (int)(Maths.Lerp(e.X, destRect.Left, destRect.Right) * image.Size.Width);
                        int py = (int)(Maths.Lerp(e.Y, destRect.Top, destRect.Bottom) * image.Size.Height);
                        MessageBox.Show(string.Format("Pixel {0}, {1}", px, py));
                    }
                }
            }

            if (e.Button == MouseButtons.Right) {
                p0 = p1 = e.Location;
                Invalidate();
            }
        }

        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);

            if (e.Button == MouseButtons.Right) {
                p0 = p1 = Point.Empty;
                Invalidate();
            }
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (e.Button == MouseButtons.Right) {
                p1 = e.Location;
                Invalidate();
            }
        }


        private static Rectangle ComputeBounds(Point p0, Point p1)
        {
            return Rectangle.FromLTRB(
                Math.Min(p0.X, p1.X),
                Math.Min(p0.Y, p1.Y),
                Math.Max(p0.X, p1.X),
                Math.Max(p0.Y, p1.Y));
        }
        private static RectangleF ComputeBounds(PointF p0, PointF p1)
        {
            return RectangleF.FromLTRB(
                Math.Min(p0.X, p1.X),
                Math.Min(p0.Y, p1.Y),
                Math.Max(p0.X, p1.X),
                Math.Max(p0.Y, p1.Y));
        }

        private static RectangleF ComputeInternalBounds(RectangleF bounds, float aspect)
        {
            RectangleF r = RectangleF.Empty;

            if (aspect > (bounds.Width / bounds.Height)) {
                r.Size = new SizeF(
                    bounds.Width, 
                    bounds.Width / aspect);
            } else {
                r.Size = new SizeF(
                    bounds.Height * aspect, 
                    bounds.Height);
            }

            r.Location = new PointF(
                bounds.X + (bounds.Width - r.Size.Width) / 2f,
                bounds.Y + (bounds.Height - r.Size.Height) / 2f);

            return r;
        }
        private static RectangleF ComputeExternalBounds(RectangleF bounds, float aspect)
        {
            RectangleF r = RectangleF.Empty;

            if (aspect < (bounds.Width / bounds.Height)) {
                r.Size = new SizeF(
                    bounds.Width, 
                    bounds.Width / aspect);
            } else {
                r.Size = new SizeF(
                    bounds.Height * aspect, 
                    bounds.Height);
            }

            r.Location = new PointF(
                bounds.X + (bounds.Width - r.Size.Width) / 2f,
                bounds.Y + (bounds.Height - r.Size.Height) / 2f);

            return r;
        }
    }
}
