using System.Runtime.InteropServices;

namespace System.Drawing.Imaging
{
    public class FastPixel : IDisposable
    {
        private Bitmap bitmap;
        private int width;
        private int height;
        private bool locked;
        private BitmapData bmpData;
        private IntPtr bmpPtr;
        private byte[] bytes;

        /// <summary>
        /// Default Constructor.
        /// </summary>
        /// <param name="bitmap">Bitmap to edit.</param>
        public FastPixel(Bitmap bitmap)
        {
            if (bitmap.PixelFormat != PixelFormat.Format32bppArgb)
                throw new ArgumentException("Bitmap must have the PixelFormat = Format32bppArgb");

            this.bitmap = bitmap;
            width = bitmap.Width;
            height = bitmap.Height;
            locked = false;
        }

        /// <summary>
        /// Locks the bitmap ready for pixel editing.
        /// </summary>
        public void Lock()
        {
            if (locked)
                throw new Exception("Bitmap already locked.");

            bmpData = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.ReadWrite, bitmap.PixelFormat);
            bmpPtr = bmpData.Scan0;
            bytes = new byte[width * height * 4];

            Marshal.Copy(bmpPtr, bytes, 0, bytes.Length);

            locked = true;
        }

        /// <summary>
        /// Unlocks the bitmap from pixel editing, and saves to the bitmap.
        /// </summary>
        /// <param name="setPixels">True to save the changes, False to discard the changes.</param>
        public void Unlock(bool setPixels)
        {
            if (!locked)
                throw new Exception("Bitmap has not been locked.");

            if (setPixels)
                Marshal.Copy(bytes, 0, bmpPtr, bytes.Length);

            bitmap.UnlockBits(bmpData);

            locked = false;
        }

        /// <summary>
        /// Sets a pixel at the given location to a given colors.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="color"></param>
        public void SetPixel(int x, int y, Color color)
        {
            if (!locked)
                throw new Exception("Bitmap has not been locked.");

            int i = (y * width + x) * 4;
            bytes[i++] = color.B;
            bytes[i++] = color.G;
            bytes[i++] = color.R;
            bytes[i++] = color.A;
        }

        /// <summary>
        /// Gets the colors of the pixel at the given location.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public Color GetPixel(int x, int y)
        {
            if (!locked)
                throw new Exception("Bitmap has not been locked.");

            int i = (y * width + x) * 4;
            int b = bytes[i++];
            int g = bytes[i++];
            int r = bytes[i++];
            int a = bytes[i++];

            return Color.FromArgb(a, r, g, b);
        }

        /// <summary>
        /// Cleanup. *Note: Does not call 'Dispose()' on the bitmap. 
        /// </summary>
        public void Dispose()
        {
            if (locked)
                Unlock(false);

            bitmap = null;
            width = 0;
            height = 0;
            locked = false;
            bmpData = null;
            bmpPtr = IntPtr.Zero;
            bytes = null;
        }
    }
}